@@ -6,8 +6,8 @@
  * Date        : 2006-12-09
  * Description : a tread-safe dcraw program interface
  *
- * Copyright (C) 2006-2008 by Gilles Caulier <caulier dot gilles at gmail dot com> 
- * Copyright (C) 2006-2008 by Marcel Wiesweg <marcel dot wiesweg at gmx dot de>
+ * Copyright (C) 2006-2009 by Gilles Caulier <caulier dot gilles at gmail dot com> 
+ * Copyright (C) 2006-2009 by Marcel Wiesweg <marcel dot wiesweg at gmx dot de>
  * Copyright (C) 2007-2008 by Guillaume Castagnino <casta at xwing dot info>
  *
  * NOTE: Do not use kdDebug() in this implementation because 
@@ -168,9 +168,12 @@ bool KDcraw::loadHalfPreview(QImage& image, const QString& path)
     qDebug("Try to use reduced RAW picture extraction");
 
     LibRaw raw;
-    raw.imgdata.params.use_auto_wb   = 1; // Use automatic white balance.
-    raw.imgdata.params.use_camera_wb = 1; // Use camera white balance, if possible.
-    raw.imgdata.params.half_size     = 1; // Half-size color image (3x faster than -q).
+    raw.imgdata.params.use_auto_wb    = 1; // Use automatic white balance.
+    raw.imgdata.params.use_camera_wb  = 1; // Use camera white balance, if possible.
+    raw.imgdata.params.half_size      = 1; // Half-size color image (3x faster than -q).
+
+    // NOTE: new magic option introduced by LibRaw 0.7.0 to to make better noise filtration.
+    raw.imgdata.params.filtering_mode = LIBRAW_FILTERING_AUTOMATIC;
 
     int ret = raw.open_file((const char*)(QFile::encodeName(path)));
     if (ret != LIBRAW_SUCCESS)
@@ -383,6 +386,9 @@ bool KDcraw::loadFromDcraw(const QString& filePath, QByteArray &imageData,
     QByteArray cameraProfile = QFile::encodeName(m_rawDecodingSettings.inputProfile);
     QByteArray outputProfile = QFile::encodeName(m_rawDecodingSettings.outputProfile);
 
+    // NOTE: new magic option introduced by LibRaw 0.7.0 to to make better noise/etc filtration.
+    raw.imgdata.params.filtering_mode = LIBRAW_FILTERING_AUTOMATIC;
+
     if (m_rawDecodingSettings.gamma16bit)
     {
         // 16 bits color depth auto-gamma is not implemented in dcraw.
@@ -742,7 +748,7 @@ QStringList KDcraw::supportedCamera()
 
 QString KDcraw::librawVersion()
 {
-    return QString(LIBRAW_VERSION_STR);
+    return QString(LIBRAW_VERSION_STR).remove("-Release");
 }
 
 }  // namespace KDcrawIface
