@@ -1,6 +1,6 @@
 /* -*- C++ -*-
  * File: libraw_cxx.cpp
- * Copyright 2008 Alex Tutubalin <lexa@lexa.ru>
+ * Copyright 2008-2009 Alex Tutubalin <lexa@lexa.ru>
  * Created: Sat Mar  8 , 2008
  *
  * LibRaw C++ interface (implementation)
@@ -23,15 +23,15 @@ extern "C"
 #endif
     void default_memory_callback(void *,const char *file,const char *where)
     {
-        fprintf (stderr,"%s: Out of memory in %s\n", file, where);
+        fprintf (stderr,"%s: Out of memory in %s\n", file?file:"unknown file", where);
     }
 
     void default_data_callback(void*,const char *file, const int offset)
     {
         if(offset < 0)
-            fprintf (stderr,"%s: Unexpected end of file\n", file);
+            fprintf (stderr,"%s: Unexpected end of file\n", file?file:"unknown file");
         else
-            fprintf (stderr,"%s: data corrupted at %d\n",file,offset); 
+            fprintf (stderr,"%s: data corrupted at %d\n",file?file:"unknown file",offset); 
     }
     const char *libraw_strerror(int e)
     {
@@ -52,6 +52,8 @@ extern "C"
                 return "No thumbnail in file";
             case    LIBRAW_UNSUPPORTED_THUMBNAIL:
                 return "Unsupported thumbnail format";
+            case LIBRAW_CANNOT_ADDMASK:
+                return "Cannot add masked pixels to resized image";
             case    LIBRAW_UNSUFFICIENT_MEMORY:
                 return "Unsufficient memory";
             case    LIBRAW_DATA_ERROR:
@@ -83,6 +85,7 @@ const float LibRaw_constants::d65_white[3] =  { 0.950456, 1, 1.088754 };
 #define S imgdata.sizes
 #define O imgdata.params
 #define C imgdata.color
+#define M imgdata.masked_pixels
 #define T imgdata.thumbnail
 #define IO libraw_internal_data.internal_output_params
 #define ID libraw_internal_data.internal_data
@@ -102,40 +105,489 @@ const float LibRaw_constants::d65_white[3] =  { 0.950456, 1, 1.088754 };
             case LIBRAW_EXCEPTION_IO_CORRUPT:           \
                 recycle();                              \
                 return LIBRAW_IO_ERROR;                 \
-            case LIBRAW_EXCEPTION_CANCELLED_BY_CALLBACK: \
+            case LIBRAW_EXCEPTION_CANCELLED_BY_CALLBACK:\
                 recycle();                              \
-                return LIBRAW_CANCELLED_BY_CALLBACK;     \
+                return LIBRAW_CANCELLED_BY_CALLBACK;    \
             default:                                    \
                 return LIBRAW_UNSPECIFIED_ERROR;        \
             } \
     }while(0)
 
+void LibRaw::derror()
+{
+    if (!libraw_internal_data.unpacker_data.data_error && libraw_internal_data.internal_data.input) 
+        {
+            if (libraw_internal_data.internal_data.input->eof())
+                {
+                    if(callbacks.data_cb)(*callbacks.data_cb)(callbacks.datacb_data,
+                                                              libraw_internal_data.internal_data.input->fname(),-1);
+                    throw LIBRAW_EXCEPTION_IO_EOF;
+                }
+            else
+                {
+                    if(callbacks.data_cb)(*callbacks.data_cb)(callbacks.datacb_data,
+                                                              libraw_internal_data.internal_data.input->fname(),
+                                                              libraw_internal_data.internal_data.input->tell());
+                    throw LIBRAW_EXCEPTION_IO_CORRUPT;
+                }
+        }
+    libraw_internal_data.unpacker_data.data_error = 1;
+}
+LibRaw:: LibRaw(unsigned int flags)
+{
+    double aber[4] = {1,1,1,1};
+    double gamm[5] = { 0.45,4.5,0,0,0 };
+    unsigned greybox[4] =  { 0, 0, UINT_MAX, UINT_MAX };
+#ifdef DCRAW_VERBOSE
+    verbose = 1;
+#else
+    verbose = 0;
+#endif
+    bzero(&imgdata,sizeof(imgdata));
+    bzero(&libraw_internal_data,sizeof(libraw_internal_data));
+    bzero(&callbacks,sizeof(callbacks));
+    callbacks.mem_cb = (flags & LIBRAW_OPIONS_NO_MEMERR_CALLBACK) ? NULL:  &default_memory_callback;
+    callbacks.data_cb = (flags & LIBRAW_OPIONS_NO_DATAERR_CALLBACK)? NULL : &default_data_callback;
+    memmove(&imgdata.params.aber,&aber,sizeof(aber));
+    memmove(&imgdata.params.gamm,&gamm,sizeof(gamm));
+    memmove(&imgdata.params.greybox,&greybox,sizeof(greybox));
+    
+    imgdata.params.bright=1;
+    imgdata.params.use_camera_matrix=-1;
+    imgdata.params.user_flip=-1;
+    imgdata.params.user_black=-1;
+    imgdata.params.user_sat=-1;
+    imgdata.params.user_qual=-1;
+    imgdata.params.output_color=1;
+    imgdata.params.output_bps=8;
+    imgdata.params.use_fuji_rotate=1;
+    imgdata.params.auto_bright_thr = 0.01;
+    imgdata.parent_class = this;
+    imgdata.progress_flags = 0;
+    tls = new LibRaw_TLS;
+    tls->init();
+}
+
+
+void* LibRaw:: malloc(size_t t)
+{
+    void *p = memmgr.malloc(t);
+    return p;
+}
+void* LibRaw::       calloc(size_t n,size_t t)
+{
+    void *p = memmgr.calloc(n,t);
+    return p;
+}
+void  LibRaw::      free(void *p)
+{
+    memmgr.free(p);
+}
+
+
+int LibRaw:: fc (int row, int col)
+{
+    static const char filter[16][16] =
+        { { 2,1,1,3,2,3,2,0,3,2,3,0,1,2,1,0 },
+          { 0,3,0,2,0,1,3,1,0,1,1,2,0,3,3,2 },
+          { 2,3,3,2,3,1,1,3,3,1,2,1,2,0,0,3 },
+          { 0,1,0,1,0,2,0,2,2,0,3,0,1,3,2,1 },
+          { 3,1,1,2,0,1,0,2,1,3,1,3,0,1,3,0 },
+          { 2,0,0,3,3,2,3,1,2,0,2,0,3,2,2,1 },
+          { 2,3,3,1,2,1,2,1,2,1,1,2,3,0,0,1 },
+          { 1,0,0,2,3,0,0,3,0,3,0,3,2,1,2,3 },
+          { 2,3,3,1,1,2,1,0,3,2,3,0,2,3,1,3 },
+          { 1,0,2,0,3,0,3,2,0,1,1,2,0,1,0,2 },
+          { 0,1,1,3,3,2,2,1,1,3,3,0,2,1,3,2 },
+          { 2,3,2,0,0,1,3,0,2,0,1,2,3,0,1,0 },
+          { 1,3,1,2,3,2,3,2,0,2,0,1,1,0,3,0 },
+          { 0,2,0,3,1,0,0,1,1,3,3,2,3,2,2,1 },
+          { 2,1,3,2,3,1,2,1,0,3,0,2,0,2,0,2 },
+          { 0,3,1,0,0,2,0,3,2,1,3,1,1,3,1,3 } };
+    
+    if (imgdata.idata.filters != 1) return FC(row,col);
+    return filter[(row+imgdata.sizes.top_margin) & 15][(col+imgdata.sizes.left_margin) & 15];
+}
+
+void LibRaw:: recycle() 
+{
+    if(libraw_internal_data.internal_data.input && libraw_internal_data.internal_data.input_internal) 
+        { 
+            delete libraw_internal_data.internal_data.input; 
+            libraw_internal_data.internal_data.input = NULL;
+        }
+    libraw_internal_data.internal_data.input_internal = 0;
+#define FREE(a) do { if(a) { free(a); a = NULL;} }while(0)
+            
+    FREE(imgdata.image); 
+    FREE(imgdata.thumbnail.thumb);
+    FREE(libraw_internal_data.internal_data.meta_data);
+    FREE(libraw_internal_data.output_data.histogram);
+    FREE(libraw_internal_data.output_data.oprof);
+    FREE(imgdata.color.profile);
+    FREE(imgdata.masked_pixels.buffer);
+    FREE(imgdata.masked_pixels.ph1_black);
+#undef FREE
+#define ZERO(a) bzero(&a,sizeof(a))
+    ZERO(imgdata.masked_pixels);
+    ZERO(imgdata.sizes);
+    ZERO(libraw_internal_data.internal_output_params);
+#undef ZERO
+    memmgr.cleanup();
+    imgdata.thumbnail.tformat = LIBRAW_THUMBNAIL_UNKNOWN;
+    imgdata.progress_flags = 0;
+    
+    tls->init();
+}
+
+const char * LibRaw::unpack_function_name()
+{
+    if(!load_raw) return "Function not set";
+
+    // sorted names order
+    if (load_raw == &LibRaw::adobe_dng_load_raw_lj)     return "adobe_dng_load_raw_lj()";
+    if (load_raw == &LibRaw::adobe_dng_load_raw_nc)     return "adobe_dng_load_raw_nc()";
+    if (load_raw == &LibRaw::canon_600_load_raw)        return "canon_600_load_raw()";
+
+    if (load_raw == &LibRaw::canon_a5_load_raw)         return "canon_a5_load_raw()";
+    if (load_raw == &LibRaw::canon_compressed_load_raw) return "canon_compressed_load_raw()";
+    if (load_raw == &LibRaw::canon_sraw_load_raw)       return "canon_sraw_load_raw()";
+
+    if (load_raw == &LibRaw::casio_qv5700_load_raw )    return "casio_qv5700_load_raw()";
+    if (load_raw == &LibRaw::eight_bit_load_raw )       return "eight_bit_load_raw()";
+    if (load_raw == &LibRaw::foveon_load_raw )          return "foveon_load_raw()";
+    if (load_raw == &LibRaw::fuji_load_raw )            return "fuji_load_raw()";
+    // 10
+    if (load_raw == &LibRaw::hasselblad_load_raw )      return "hasselblad_load_raw()";
+    if (load_raw == &LibRaw::imacon_full_load_raw )     return "imacon_full_load_raw()";
+    if (load_raw == &LibRaw::kodak_262_load_raw )       return "kodak_262_load_raw()";
+
+    if (load_raw == &LibRaw::kodak_65000_load_raw )     return "kodak_65000_load_raw()";
+    if (load_raw == &LibRaw::kodak_dc120_load_raw )     return "kodak_dc120_load_raw()";
+    if (load_raw == &LibRaw::kodak_jpeg_load_raw )      return "kodak_jpeg_load_raw()";
+
+    if (load_raw == &LibRaw::kodak_radc_load_raw )      return "kodak_radc_load_raw()";
+    if (load_raw == &LibRaw::kodak_rgb_load_raw )       return "kodak_rgb_load_raw()";
+    if (load_raw == &LibRaw::kodak_yrgb_load_raw )      return "kodak_yrgb_load_raw()";
+    if (load_raw == &LibRaw::kodak_ycbcr_load_raw )     return "kodak_ycbcr_load_raw()";
+    // 20
+    if (load_raw == &LibRaw::leaf_hdr_load_raw )        return "leaf_hdr_load_raw()";
+    if (load_raw == &LibRaw::lossless_jpeg_load_raw)    return "lossless_jpeg_load_raw()";
+    if (load_raw == &LibRaw::minolta_rd175_load_raw )   return "minolta_rd175_load_raw()";
+
+    if (load_raw == &LibRaw::nikon_compressed_load_raw) return "nikon_compressed_load_raw()";
+    if (load_raw == &LibRaw::nikon_e900_load_raw )      return "nikon_e900_load_raw()";
+    if (load_raw == &LibRaw::nokia_load_raw )           return "nokia_load_raw()";
+
+    if (load_raw == &LibRaw::olympus_e300_load_raw )    return "olympus_e300_load_raw()";
+    if (load_raw == &LibRaw::olympus_e410_load_raw )    return "olympus_e410_load_raw()";
+    if (load_raw == &LibRaw::packed_12_load_raw )       return "packed_12_load_raw()";
+    if (load_raw == &LibRaw::panasonic_load_raw )       return "panasonic_load_raw()";
+    // 30
+    if (load_raw == &LibRaw::pentax_k10_load_raw )      return "pentax_k10_load_raw()";
+    if (load_raw == &LibRaw::phase_one_load_raw )       return "phase_one_load_raw()";
+    if (load_raw == &LibRaw::phase_one_load_raw_c )     return "phase_one_load_raw_c()";
+
+    if (load_raw == &LibRaw::quicktake_100_load_raw )   return "quicktake_100_load_raw()";
+    if (load_raw == &LibRaw::rollei_load_raw )          return "rollei_load_raw()";
+    if (load_raw == &LibRaw::sinar_4shot_load_raw )     return "sinar_4shot_load_raw()";
+
+    if (load_raw == &LibRaw::smal_v6_load_raw )         return "smal_v6_load_raw()";
+    if (load_raw == &LibRaw::smal_v9_load_raw )         return "smal_v9_load_raw()";
+    if (load_raw == &LibRaw::sony_load_raw )            return "sony_load_raw()";
+    if (load_raw == &LibRaw::sony_arw_load_raw )        return "sony_arw_load_raw()";
+    // 40
+    if (load_raw == &LibRaw::sony_arw2_load_raw )       return "sony_arw2_load_raw()";
+    if (load_raw == &LibRaw::unpacked_load_raw )        return "unpacked_load_raw()";
+    // 42 total
+        
+    return "Unknown unpack function";
+}
+
+
+void LibRaw:: merror (void *ptr, const char *where)
+{
+    if (ptr) return;
+    if(callbacks.mem_cb)(*callbacks.mem_cb)(callbacks.memcb_data,
+                                            libraw_internal_data.internal_data.input
+                                            ?libraw_internal_data.internal_data.input->fname()
+                                            :NULL,
+                                            where);
+    throw LIBRAW_EXCEPTION_ALLOC;
+}
+
+ushort * LibRaw::get_masked_pointer(int row, int col) 
+{ 
+    if(row<0 || col < 0) return NULL;
+    if(!M.buffer) return NULL; 
+    if(row < S.top_margin)
+        {
+            // top band
+            if(col < S.left_margin)
+                {
+                    return &(M.tl[row*S.left_margin+col]);
+                }
+            else if (col < S.left_margin + S.width)
+                {
+                    int icol = col - S.left_margin;
+                    return &(M.top[row*S.width+icol]);
+                }
+            else if (col < S.raw_width)
+                {
+                    int icol = col - S.left_margin - S.width;
+                    return &(M.tr[row*S.right_margin+icol]);
+                }
+            else
+                return NULL; // out of bounds
+        }
+    else if (row < S.top_margin + S.height)
+        {
+            //normal image height
+            int irow = row - S.top_margin;
+            if(col < S.left_margin)
+                {
+                    return &M.left[irow*S.left_margin + col];
+                }
+            else if (col < S.left_margin + S.width)
+                {
+                    // central image
+                    return NULL;
+                }
+            else if (col < S.raw_width)
+                {
+                    int icol = col - S.left_margin - S.width;
+                    return &M.right[irow*S.right_margin+icol];
+                }
+            else
+                return NULL; // out of bounds
+        }
+    else if (row < S.raw_height)
+        {
+            int irow = row - S.top_margin - S.height;
+            // bottom band
+            if(col < S.left_margin)
+                {
+                    return &M.bl[irow*S.left_margin+col];
+                }
+            else if (col < S.left_margin + S.width)
+                {
+                    int icol = col - S.left_margin;
+                    return &M.bottom[irow*S.width + icol];
+                }
+            else if (col < S.raw_width)
+                {
+                    int icol = col - S.left_margin - S.width;
+                    return &M.br[irow*S.right_margin + icol];
+                }
+            else
+                return NULL; // out of bounds
+        }
+    else
+        {
+            // out of bounds
+            return NULL;
+        }
+    // fallback
+    return NULL;
+}
+
+void LibRaw:: init_masked_ptrs()
+{
+    if(!M.buffer) return;
+    
+    // top band
+    M.tl = M.buffer;
+    M.top = M.tl +(S.top_margin*S.left_margin);
+    M.tr =  M.top + (S.top_margin*S.width);
+    
+    // left-right
+    M.left = M.tr + (S.top_margin * S.right_margin);
+    M.right = M.left + (S.left_margin * S.height);
+
+    // bottom band
+    M.bl = M.right + (S.right_margin * S.height);
+    M.bottom = M.bl + (S.left_margin * S.bottom_margin);
+    M.br = M.bottom + (S.width * S.bottom_margin);
+
+}
+
+int LibRaw::add_masked_borders_to_bitmap()
+{
+    CHECK_ORDER_HIGH(LIBRAW_PROGRESS_PRE_INTERPOLATE);
+    CHECK_ORDER_LOW(LIBRAW_PROGRESS_LOAD_RAW);
+
+    if(S.width != S.iwidth || S.height!=S.iheight)
+        return LIBRAW_CANNOT_ADDMASK;
+
+    if(P1.is_foveon || !P1.filters)
+        return LIBRAW_CANNOT_ADDMASK;
+        
+    if(!imgdata.image)
+        return LIBRAW_OUT_OF_ORDER_CALL;
+
+    if(S.raw_width < S.width || S.raw_height < S.height)
+        return LIBRAW_SUCCESS; // nothing to do or already called
+
+    if(S.width == S.raw_width && S.height == S.raw_height)
+        return LIBRAW_SUCCESS; // nothing to do or already called
+
+    ushort (*newimage)[4];
+
+    newimage = (ushort (*)[4]) calloc (S.raw_height*S.raw_width, sizeof (*newimage));
+    merror (newimage, "add_masked_borders_to_bitmap()");
+
+    int r,c;
+    // top rows
+    for (r=0; r<S.top_margin;r++)
+        for(c=0;c<S.raw_width;c++)
+            {
+                ushort *p = get_masked_pointer(r,c);
+                if(p)
+                    newimage[r*S.raw_width+c][FC(r,c)] = *p;
+            }
+    // middle rows
+    for (r=S.top_margin; r<S.top_margin+S.height;r++)
+        {
+            int row = r-S.top_margin;
+            for(c=0;c<S.left_margin;c++)
+                {
+                    ushort *p = get_masked_pointer(r,c);
+                    if(p)
+                        newimage[r*S.raw_width+c][FC(r,c)] =  *p;
+                }
+            for(c=S.left_margin; c<S.left_margin+S.iwidth;c++)
+                {
+                    int col = c - S.left_margin;
+                    newimage[r*S.raw_width+c][FC(r,c)] = imgdata.image[row*S.iwidth+col][FC(row,col)];
+                }
+            for(c=S.left_margin+S.iwidth;c<S.raw_width;c++)
+                {
+                    ushort *p = get_masked_pointer(r,c);
+                    if(p)
+                        newimage[r*S.raw_width+c][FC(r,c)] =  *p;
+                }
+        }
+    // bottom rows
+    for (r=S.top_margin+S.height; r<S.raw_height;r++)
+        for(c=0;c<S.raw_width;c++)
+            {
+                ushort *p = get_masked_pointer(r,c);
+                if(p)
+                    newimage[r*S.raw_width+c][FC(r,c)] = *p;
+            }
+    free(imgdata.image);
+    imgdata.image=newimage;
+    S.iwidth = S.width = S.raw_width;
+    S.iheight = S.height = S.raw_height;
+    return LIBRAW_SUCCESS;
+}
 
 int LibRaw::open_file(const char *fname)
 {
+    // this stream will close on recycle()
+    LibRaw_file_datastream *stream = new LibRaw_file_datastream(fname);
+    if(!stream->valid())
+        {
+            delete stream;
+            return LIBRAW_IO_ERROR;
+        }
+    ID.input_internal = 0; // preserve from deletion on error
+    int ret = open_datastream(stream);
+    if (ret == LIBRAW_SUCCESS)
+        {
+            ID.input_internal =1 ; // flag to delete datastream on recycle
+        }
+    else
+        {
+            delete stream;
+            ID.input_internal = 0;
+        }
+    return ret;
+}
 
-    if(!fname)
+int LibRaw::open_buffer(void *buffer, size_t size)
+{
+    // this stream will close on recycle()
+    if(!buffer  || buffer==(void*)-1)
+        return LIBRAW_IO_ERROR;
+
+    LibRaw_buffer_datastream *stream = new LibRaw_buffer_datastream(buffer,size);
+    if(!stream->valid())
+        {
+            delete stream;
+            return LIBRAW_IO_ERROR;
+        }
+    ID.input_internal = 0; // preserve from deletion on error
+    int ret = open_datastream(stream);
+    if (ret == LIBRAW_SUCCESS)
+        {
+            ID.input_internal =1 ; // flag to delete datastream on recycle
+        }
+    else
+        {
+            delete stream;
+            ID.input_internal = 0;
+        }
+    return ret;
+}
+
+
+int LibRaw::open_datastream(LibRaw_abstract_datastream *stream)
+{
+
+    if(!stream)
         return ENOENT;
-        
+    if(!stream->valid())
+        return LIBRAW_IO_ERROR;
     recycle();
-    libraw_internal_data.internal_data.ifname = (char*)fname;
-    try {
-        if(!(ID.input = fopen(ID.ifname,"rb")))
-            return errno;
 
+    try {
+        ID.input = stream;
         SET_PROC_FLAG(LIBRAW_PROGRESS_OPEN);
 
         if (O.use_camera_matrix < 0)
             O.use_camera_matrix = O.use_camera_wb;
 
         identify();
+
+        if(IO.fuji_width)
+            {
+                IO.fwidth = S.width;
+                IO.fheight = S.height;
+                S.iwidth = S.width = IO.fuji_width << !libraw_internal_data.unpacker_data.fuji_layout;
+                S.iheight = S.height = S.raw_height;
+                S.raw_height += 2*S.top_margin;
+            }
+
+        int saved_raw_width = S.raw_width;
+        int saved_width = S.width;
+        // from packed_12_load_raw
+        if ((load_raw == &LibRaw:: packed_12_load_raw) && (S.raw_width * 2 >= S.width * 3))
+            {	
+                // raw_width is in bytes!
+                S.raw_width = S.raw_width * 2 / 3;	
+            }
+        else if (S.pixel_aspect < 0.95 || S.pixel_aspect > 1.05)
+            {
+                S.width*=S.pixel_aspect;
+            }
+
+        if(S.raw_width>S.width + S.left_margin)
+            S.right_margin = S.raw_width - S.width - S.left_margin;
+
+        if(S.raw_height > S.height + S.top_margin)
+            S.bottom_margin = S.raw_height - S.height - S.top_margin;
+
+        S.raw_width = saved_raw_width;
+        S.width = saved_width;
+
         if(C.profile_length)
             {
                 if(C.profile) free(C.profile);
                 C.profile = malloc(C.profile_length);
                 merror(C.profile,"LibRaw::open_file()");
-                fseek(ID.input,ID.profile_offset,SEEK_SET);
-                fread(C.profile,C.profile_length,1,ID.input);
+                ID.input->seek(ID.profile_offset,SEEK_SET);
+                ID.input->read(C.profile,C.profile_length,1);
             }
         
         SET_PROC_FLAG(LIBRAW_PROGRESS_IDENTIFY);
@@ -171,6 +623,7 @@ int LibRaw::open_file(const char *fname)
     
     SET_PROC_FLAG(LIBRAW_PROGRESS_SIZE_ADJUST);
 
+
     return LIBRAW_SUCCESS;
 }
 
@@ -197,21 +650,37 @@ int LibRaw::unpack(void)
         
         imgdata.image = (ushort (*)[4]) calloc (S.iheight*S.iwidth, sizeof (*imgdata.image));
         merror (imgdata.image, "unpack()");
+
+
+        if(S.top_margin || S.left_margin || S.right_margin || S.bottom_margin)
+            {
+                unsigned sz = S.raw_height*(S.left_margin+S.right_margin) 
+                    + S.width*(S.top_margin+S.bottom_margin);
+                imgdata.masked_pixels.buffer = (ushort*) calloc(sz, sizeof(ushort)); 
+                merror (imgdata.masked_pixels.buffer, "unpack()");
+                init_masked_ptrs();
+            }
         if (libraw_internal_data.unpacker_data.meta_length) 
             {
                 libraw_internal_data.internal_data.meta_data = 
                     (char *) malloc (libraw_internal_data.unpacker_data.meta_length);
                 merror (libraw_internal_data.internal_data.meta_data, "LibRaw::unpack()");
             }
-        fseek (ID.input, libraw_internal_data.unpacker_data.data_offset, SEEK_SET);
+        ID.input->seek(libraw_internal_data.unpacker_data.data_offset, SEEK_SET);
         // foveon_load_raw produces different data for document_mode, we'll
         // deal with it in dcraw_document_mode_processing
         int save_document_mode = O.document_mode;
         O.document_mode = 0;
+
+        if(!own_filtering_supported() && (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT))
+            O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC_BIT; // turn on black and zeroes filtering
         
         (this->*load_raw)();
         
         O.document_mode = save_document_mode;
+
+        if (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT)
+            O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC; // restore automated mode
         
         SET_PROC_FLAG(LIBRAW_PROGRESS_LOAD_RAW);
         RUN_CALLBACK(LIBRAW_PROGRESS_LOAD_RAW,1,2);
@@ -229,6 +698,13 @@ int LibRaw::dcraw_document_mode_processing(void)
     CHECK_ORDER_LOW(LIBRAW_PROGRESS_LOAD_RAW);
 
     try {
+
+        if(IO.fwidth) 
+            rotate_fuji_raw();
+
+        if(!own_filtering_supported() && (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT))
+            O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC_BIT; // turn on black and zeroes filtering
+
         O.document_mode = 2;
         if(P1.is_foveon)
             {
@@ -243,7 +719,7 @@ int LibRaw::dcraw_document_mode_processing(void)
             }
 
         O.use_fuji_rotate = 0;
-        if (IO.zero_is_bad)
+        if (!(O.filtering_mode & LIBRAW_FILTERING_NOZEROES) && IO.zero_is_bad)
             {
                 remove_zeroes();
                 SET_PROC_FLAG(LIBRAW_PROGRESS_REMOVE_ZEROES);
@@ -258,7 +734,9 @@ int LibRaw::dcraw_document_mode_processing(void)
                 subtract (O.dark_frame);
                 SET_PROC_FLAG(LIBRAW_PROGRESS_DARK_FRAME);
             }
-        
+        if(O.filtering_mode & LIBRAW_FILTERING_NOBLACKS)
+            C.black=0;
+
         if (O.user_black >= 0) 
             C.black = O.user_black;
 
@@ -308,6 +786,10 @@ int LibRaw::dcraw_document_mode_processing(void)
         if (O.use_fuji_rotate)
             stretch();
         SET_PROC_FLAG(LIBRAW_PROGRESS_STRETCH);
+
+        if (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT)
+            O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC; // restore automated mode
+
         return 0;
     }
     catch ( LibRaw_exceptions err) {
@@ -423,7 +905,7 @@ libraw_processed_image_t *LibRaw::dcraw_make_mem_image(int *errcode)
         {
             libraw_internal_data.output_data.histogram = 
                 (int (*)[LIBRAW_HISTOGRAM_SIZE]) malloc(sizeof(*libraw_internal_data.output_data.histogram)*4);
-            merror(libraw_internal_data.output_data.histogram,"LibRaw::dcraw_ppm_tiff_writer()");
+            merror(libraw_internal_data.output_data.histogram,"LibRaw::dcraw_make_mem_image()");
         }
 
     unsigned ds = S.height * S.width * (O.output_bps/8) * P1.colors;
@@ -552,7 +1034,7 @@ void LibRaw::kodak_thumb_loader()
     imgdata.image = (ushort (*)[4]) calloc (S.iheight*S.iwidth, sizeof (*imgdata.image));
     merror (imgdata.image, "LibRaw::kodak_thumb_loader()");
 
-    fseek (ID.input,ID.toffset, SEEK_SET);
+    ID.input->seek(ID.toffset, SEEK_SET);
     // read kodak thumbnail into T.image[]
     (this->*thumb_load_raw)();
 
@@ -700,7 +1182,7 @@ void LibRaw::foveon_thumb_loader (void)
             merror (buf, "foveon_thumb()");
             for (row=0; row < T.theight; row++) 
                 {
-                    fread(buf, 1, bwide, ID.input);
+                    ID.input->read(buf, 1, bwide);
                     memmove(T.thumb+(row*T.twidth*3),buf,T.twidth*3);
                 }
             free(buf);
@@ -725,7 +1207,7 @@ void LibRaw::foveon_thumb_loader (void)
                                     {
                                         if ((bit = (bit-1) & 31) == 31)
                                             for (i=0; i < 4; i++)
-                                                bitbuf = (bitbuf << 8) + fgetc(ID.input);
+                                                bitbuf = (bitbuf << 8) + ID.input->get_char();
                                         dindex = dindex->branch[bitbuf >> bit & 1];
                                     }
                                 pred[c] += dindex->leaf;
@@ -759,13 +1241,13 @@ int LibRaw::unpack_thumb(void)
             } 
         else 
             {
-                fseek (ID.input,ID.toffset, SEEK_SET);
+                ID.input->seek(ID.toffset, SEEK_SET);
                 if ( write_thumb == &LibRaw::jpeg_thumb)
                     {
                         if(T.thumb) free(T.thumb);
                         T.thumb = (char *) malloc (T.tlength);
                         merror (T.thumb, "jpeg_thumb()");
-                        fread (T.thumb, 1, T.tlength, ID.input);
+                        ID.input->read (T.thumb, 1, T.tlength);
                         T.tcolors = 3;
                         T.tformat = LIBRAW_THUMBNAIL_JPEG;
                         SET_PROC_FLAG(LIBRAW_PROGRESS_THUMB_LOAD);
@@ -779,7 +1261,7 @@ int LibRaw::unpack_thumb(void)
                         T.thumb = (char *) malloc (T.tlength);
                         merror (T.thumb, "ppm_thumb()");
 
-                        fread (T.thumb, 1, T.tlength, ID.input);
+                        ID.input->read(T.thumb, 1, T.tlength);
 
                         T.tformat = LIBRAW_THUMBNAIL_BITMAP;
                         SET_PROC_FLAG(LIBRAW_PROGRESS_THUMB_LOAD);
@@ -822,7 +1304,10 @@ int LibRaw::dcraw_thumb_writer(const char *fname)
         return errno;
 
     if(!T.thumb)
-        return LIBRAW_OUT_OF_ORDER_CALL;
+	{
+		fclose(tfp);
+        	return LIBRAW_OUT_OF_ORDER_CALL;
+	}
 
     try {
         switch (T.tformat)
@@ -855,6 +1340,17 @@ int LibRaw::adjust_sizes_info_only(void)
         {
             if (IO.fuji_width) 
                 {
+                    // restore saved values
+                    if(IO.fheight)
+                        {
+                            S.height = IO.fheight;
+                            S.width = IO.fwidth;
+                            S.iheight = (S.height + IO.shrink) >> IO.shrink;
+                            S.iwidth  = (S.width  + IO.shrink) >> IO.shrink;
+                            S.raw_height -= 2*S.top_margin;
+                            IO.fheight = IO.fwidth = 0; // prevent repeated calls
+                        }
+                    // dcraw code
                     IO.fuji_width = (IO.fuji_width - 1 + IO.shrink) >> IO.shrink;
                     S.iwidth = (ushort)(IO.fuji_width / sqrt(0.5));
                     S.iheight = (ushort)( (S.iheight - IO.fuji_width) / sqrt(0.5));
@@ -876,6 +1372,53 @@ int LibRaw::adjust_sizes_info_only(void)
     return 0;
 }
 
+int LibRaw::rotate_fuji_raw(void)
+{
+    CHECK_ORDER_LOW(LIBRAW_PROGRESS_LOAD_RAW);
+    CHECK_ORDER_HIGH(LIBRAW_PROGRESS_PRE_INTERPOLATE);
+
+
+    if(!IO.fwidth) return LIBRAW_SUCCESS;
+    int row,col,r,c;
+    ushort (*newimage)[4];
+    ushort fiwidth,fiheight;
+
+    fiheight = (IO.fheight + IO.shrink) >> IO.shrink;
+    fiwidth = (IO.fwidth + IO.shrink) >> IO.shrink;
+    
+    newimage = (ushort (*)[4]) calloc (fiheight*fiwidth, sizeof (*newimage));
+    merror(newimage,"rotate_fuji_raw()");
+    for(row=0;row<S.height;row++)
+        {
+            for(col=0;col<S.width;col++)
+                {
+
+                    if (libraw_internal_data.unpacker_data.fuji_layout) {
+                        r = IO.fuji_width - 1 - col + (row >> 1);
+                        c = col + ((row+1) >> 1);
+                    } else {
+                        r = IO.fuji_width - 1 + row - (col >> 1);
+                        c = row + ((col+1) >> 1);
+                    }
+                    newimage[((r) >> IO.shrink)*fiwidth + ((c) >> IO.shrink)][FC(r,c)] = 
+                        imgdata.image[((row) >> IO.shrink)*S.iwidth + ((col) >> IO.shrink)][FC(r,c)];
+                }
+        }
+    // restore fuji sizes!
+    S.height = IO.fheight;
+    S.width = IO.fwidth;
+    S.iheight = (S.height + IO.shrink) >> IO.shrink;
+    S.iwidth  = (S.width  + IO.shrink) >> IO.shrink;
+    S.raw_height -= 2*S.top_margin;
+    IO.fheight = IO.fwidth = 0; // prevent repeated calls
+
+    free(imgdata.image);
+    imgdata.image = newimage;
+    return LIBRAW_SUCCESS;
+    
+}
+
+
 int LibRaw::dcraw_process(void)
 {
     int quality,i;
@@ -885,10 +1428,18 @@ int LibRaw::dcraw_process(void)
     CHECK_ORDER_HIGH(LIBRAW_PROGRESS_PRE_INTERPOLATE);
 
     try {
+
+        if(IO.fwidth) 
+            rotate_fuji_raw();
+
+
+        if(!own_filtering_supported() && (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT))
+            O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC_BIT; // turn on black and zeroes filtering
+
         if(O.half_size) 
             O.four_color_rgb = 1;
 
-        if (IO.zero_is_bad) 
+        if (!(O.filtering_mode & LIBRAW_FILTERING_NOZEROES) && IO.zero_is_bad) 
             {
                 remove_zeroes();
                 SET_PROC_FLAG(LIBRAW_PROGRESS_REMOVE_ZEROES);
@@ -905,6 +1456,10 @@ int LibRaw::dcraw_process(void)
             }
 
         quality = 2 + !IO.fuji_width;
+
+        if(O.filtering_mode & LIBRAW_FILTERING_NOBLACKS)
+            C.black=0;
+
         if (O.user_qual >= 0) quality = O.user_qual;
         if (O.user_black >= 0) C.black = O.user_black;
         if (O.user_sat > 0) C.maximum = O.user_sat;
@@ -990,6 +1545,8 @@ int LibRaw::dcraw_process(void)
                 stretch();
                 SET_PROC_FLAG(LIBRAW_PROGRESS_STRETCH);
             }
+        if (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT)
+            O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC; // restore automated mode
         return 0;
     }
     catch ( LibRaw_exceptions err) {
@@ -1105,6 +1662,7 @@ static const char  *static_camera_list[] =
 "Fuji IS-1",
 "Hasselblad CFV",
 "Hasselblad H3D",
+"Hasselblad V96C",
 "Imacon Ixpress 16-megapixel",
 "Imacon Ixpress 22-megapixel",
 "Imacon Ixpress 39-megapixel",
@@ -1195,6 +1753,7 @@ static const char  *static_camera_list[] =
 "Nikon D2X",
 "Nikon D2Xs",
 "Nikon D3",
+"Nikon D3X",
 "Nikon D40",
 "Nikon D40X",
 "Nikon D50",
@@ -1256,6 +1815,7 @@ static const char  *static_camera_list[] =
 "Olympus SP510UZ",
 "Olympus SP550UZ",
 "Olympus SP560UZ",
+"Olympus SP570UZ",
 "Panasonic DMC-FZ8",
 "Panasonic DMC-FZ18",
 "Panasonic DMC-FZ28",
@@ -1279,6 +1839,7 @@ static const char  *static_camera_list[] =
 "Pentax K100D",
 "Pentax K100D Super",
 "Pentax K200D",
+"Pentax K2000/K-m",
 "Pentax Optio S",
 "Pentax Optio S4",
 "Pentax Optio 33WR",
@@ -1297,6 +1858,7 @@ static const char  *static_camera_list[] =
 "RoverShot 3320af",
 "Samsung GX-1S",
 "Samsung GX-10",
+"Samsung S85 (hacked)",
 "Sarnoff 4096x5440",
 "Sigma SD9",
 "Sigma SD10",
